__all__ = ('InteractionEvent',)

from warnings import warn

from scarletio import Future, copy_docs, export, include, shield

from ...bases import DiscordEntity, EventBase
from ...channel import Channel, ChannelType, create_partial_channel_from_id
from ...component import ComponentType
from ...core import (
    APPLICATION_ID_TO_CLIENT, INTERACTION_EVENT_MESSAGE_WAITERS, INTERACTION_EVENT_RESPONSE_WAITERS, KOKORO
)
from ...guild.guild.guild_boost_perks import LEVEL_0
from ...localization.utils import LOCALE_DEFAULT
from ...message import Message
from ...permission import Permission
from ...precreate_helpers import process_precreate_parameters, raise_extra
from ...resolved import Resolved
from ...user import ClientUserBase, ZEROUSER, create_partial_user_from_id
from ...utils import now_as_id

from ..interaction_metadata import InteractionMetadataBase
from ..interaction_metadata.fields import (
    validate_application_command_id, validate_application_command_name, validate_component, validate_components,
    validate_custom_id, validate_options, validate_target_id, validate_target_type
)
from ..responding.constants import (
    RESPONSE_FLAG_ACKNOWLEDGED, RESPONSE_FLAG_ACKNOWLEDGING, RESPONSE_FLAG_DEFERRED, RESPONSE_FLAG_DEFERRING,
    RESPONSE_FLAG_EPHEMERAL, RESPONSE_FLAG_NONE, RESPONSE_FLAG_RESPONDED, RESPONSE_FLAG_RESPONDING
)

from .constants import DEFAULT_INTERACTION_METADATA, INTERACTION_EVENT_EXPIRE_AFTER_ID_DIFFERENCE
from .fields import (
    parse_application_id, parse_application_permissions, parse_attachment_size_limit, parse_authorizer_user_ids,
    parse_channel, parse_entitlements, parse_guild, parse_id, parse_message, parse_resolved, parse_token, parse_type,
    parse_user, parse_user_locale, parse_user_permissions, put_application_id, put_application_permissions,
    put_attachment_size_limit, put_authorizer_user_ids, put_channel, put_entitlements, put_guild, put_id, put_message,
    put_resolved, put_token, put_type, put_user, put_user_locale, put_user_permissions, validate_application_id,
    validate_application_permissions, validate_attachment_size_limit, validate_authorizer_user_ids, validate_channel,
    validate_entitlements, validate_guild, validate_id, validate_interaction, validate_message, validate_resolved,
    validate_token, validate_type, validate_user, validate_user_locale, validate_user_permissions
)
from .preinstanced import InteractionType


create_partial_guild_from_id = include('create_partial_guild_from_id')



PRECREATE_FIELDS = {
    'attachment_size_limit': ('attachment_size_limit', validate_attachment_size_limit),
    'application': ('application_id', validate_application_id),
    'application_id': ('application_id', validate_application_id),
    'application_permissions': ('application_permissions', validate_application_permissions),
    'authorizer_user_ids': ('authorizer_user_ids', validate_authorizer_user_ids),
    'channel': ('channel', validate_channel),
    'entitlements': ('entitlements', validate_entitlements),
    'guild': ('guild', validate_guild),
    'user_locale': ('user_locale', validate_user_locale),
    'message': ('message', validate_message),
    'resolved': ('resolved', validate_resolved),
    'token': ('token', validate_token),
    'user': ('user', validate_user),
    'user_permissions': ('user_permissions', validate_user_permissions),
}


def _unpack_interaction_metadata_into_keyword_parameters(interaction_type, interaction_metadata, keyword_parameters):
    if interaction_type is InteractionType.application_command:
        keyword_parameters['application_command_id'] = interaction_metadata.application_command_id
        keyword_parameters['application_command_name'] = interaction_metadata.application_command_name
        keyword_parameters['options'] = interaction_metadata.options
        keyword_parameters['target_id'] = interaction_metadata.target_id
        keyword_parameters['target_type'] = interaction_metadata.target_type
        
    elif interaction_type is InteractionType.message_component:
        keyword_parameters['component'] = interaction_metadata.component
        
    elif interaction_type is InteractionType.application_command_autocomplete:
        keyword_parameters['application_command_id'] = interaction_metadata.application_command_id
        keyword_parameters['application_command_name'] = interaction_metadata.application_command_name
        keyword_parameters['options'] = interaction_metadata.options
        
    elif interaction_type is InteractionType.form_submit:
        keyword_parameters['components'] = interaction_metadata.components
        keyword_parameters['custom_id'] = interaction_metadata.custom_id


@export
class InteractionEvent(DiscordEntity, EventBase, immortal = True):
    """
    Represents a processed `INTERACTION_CREATE` dispatch event.
    
    Attributes
    ----------
    id : `int`
        The interaction's id.
    
    _async_task : `None`, ``Task``
        Task set if interaction event is acknowledged asynchronously.
    
    _response_flags : `int`
        The response order state of ``InteractionEvent``
        
        +-------------------------------+-------+---------------------------------------------------+
        | Respective name               | Shift | Description                                       |
        +===============================+=======+===================================================+
        | RESPONSE_FLAG_DEFERRING       | 0     | The vent is being acknowledged.                   |
        +-------------------------------+-------+---------------------------------------------------+
        | RESPONSE_FLAG_DEFERRED        | 1     | The event was acknowledged and response will be   |
        |                               |       | sent later. Shows loading screen for the user.    |
        +-------------------------------+-------+---------------------------------------------------+
        | RESPONSE_FLAG_RESPONDING      | 2     | Responding to the interaction.                    |
        +-------------------------------+-------+---------------------------------------------------+
        | RESPONSE_FLAG_RESPONDED       | 3     | Response was sent on the interaction.             |
        +-------------------------------+-------+---------------------------------------------------+
        | RESPONSE_FLAG_EPHEMERAL       | 4     | Whether the main response is an ephemeral,        |
        |                               |       | showing for the invoking user only.               |
        +-------------------------------+-------+---------------------------------------------------+
        
        Can be used by extensions and is used by the the ``Client``-s to ensure correct flow order.
    
    application_id : `int`
        The interaction's application's identifier.
    
    application_permissions : ``Permission``
        The permissions granted to the application in the guild.
    
    attachment_size_limit : `int`
        The size limit in bytes for each attachment.
    
    authorizer_user_ids : `None | dict<ApplicationIntegrationType, int>`
        The users' identifier who authorized the integration.
    
    entitlements : ``None | tuple<Entitlement>``
        The applicable entitlements for the event's context.
        These can both target guild and user as well.
    
    channel : ``Channel``
        The channel from where the interaction was called.
    
    guild : ``None | Guild``
        The guild from where the interaction was called from.
    
    message : ``None | Message``
        The message from where the interaction was received. Applicable for message components.
    
    metadata : ``InteractionMetadataBase``
        Interaction metadata containing its type specific fields.
    
    resolved : ``None | Resolved``
        Contains the received entities.
    
    token : `str`
        Interaction's token used when responding on it.
    
    type : ``InteractionType``
        The interaction's type.
    
    user : ``ClientUserBase``
        The user who called the interaction.
    
    user_locale : ``Locale``
        The selected language of the invoking user.
    
    user_permissions : ``Permission``
        The user's permissions in the respective channel.
    
    Notes
    -----
    The interaction token can be used for 15 minutes, tho if it is not used within the first 3 seconds, it is
    invalidated immediately.
    
    Interaction event instances are weakreferable.
    """
    __slots__ = (
        '_async_task', '_response_flags', 'application_id', 'application_permissions', 'attachment_size_limit',
        'authorizer_user_ids', 'entitlements', 'channel', 'guild', 'message', 'metadata', 'resolved', 'token',
        'type', 'user', 'user_locale', 'user_permissions',
    )
    
    def __new__(
        cls,
        *,
        application_id = ...,
        application_permissions = ...,
        attachment_size_limit = ...,
        authorizer_user_ids = ...,
        channel = ...,
        entitlements = ...,
        guild = ...,
        guild_locale = ...,
        interaction = ...,
        interaction_type = ...,
        locale = ...,
        message = ...,
        resolved = ...,
        token = ..., 
        user = ...,
        user_locale = ...,
        user_permissions = ...,
        **keyword_parameters,
    ):
        """
        Creates a partial interaction event.
        
        Parameters
        ----------
        application_id : `int`, `str`, Optional (Keyword only)
            The interaction's application's identifier.
        
        application_permissions : ``Permission``, `int`, Optional (Keyword only)
            The permissions granted to the application in the guild.
        
        attachment_size_limit : `int`, Optional (Keyword only)
            The size limit in bytes for each attachment.
        
        authorizer_user_ids : `None | dict<ApplicationIntegrationType | int, int | ClientUserBase>` \
                , Optional (Keyword only)
            The users' identifier who authorized the integration.
        
        channel : ``Channel``, Optional (Keyword only)
            The channel from where the interaction was called.
        
        entitlements : `None`, `iterable` of ``Entitlement``, Optional (Keyword only)
            The applicable entitlements for the event's context.
        
        guild : ``None | Guild``, Optional (Keyword only)
            The guild from where the interaction was called from.
        
        message : ``None | Message``, Optional (Keyword only)
            The message from where the interaction was received.
        
        resolved : ``None | Resolved``, Optional (Keyword only)
            Contains the received entities.
        
        token : `str`, Optional (Keyword only)
            Interaction's token used when responding on it.
        
        type : ``InteractionType``, `int`, Optional (Keyword only)
            The interaction's type.
        
        user : ``ClientUserBase``, Optional (Keyword only)
            The user who called the interaction.
        
        user_locale : ``Locale``, `str`, Optional (Keyword only)
            The selected language of the invoking user.
        
        user_permissions : ``Permission``, `int`, Optional (Keyword only)
            The user's permissions in the respective channel.
        
        **keyword_parameters : Keyword parameters
            Keyword parameters defining the type specific fields of the interaction.
        
        Other Parameters
        ----------------
        application_command_id : ``None | int | ApplicationCommand``, Optional (Keyword only)
            The represented application command's identifier number.
        
        application_command_name : `None | str`, Optional (Keyword only)
            The represented application command's name.
        
        component : ``None | InteractionComponent``, Optional (Keyword only)
            The interacted component.
        
        components : ``None | iterable<InteractionComponent>``, Optional (Keyword only)
            Submitted component values of a form submit interaction.
        
        custom_id : `None | str`, Optional (Keyword only)
            Form interaction's custom identifier.
        
        options : ``None | iterable<InteractionOption>``, Optional (Keyword only)
            Application command option representations. Like sub-command or parameter.
        
        target_id : `int`, Optional (Keyword only)
            The interaction's target's identifier. Applicable for context commands.
        
        target_type : ``None | int | ApplicationCommandTargetType``, Optional (Keyword only)
            The invoked application command's target type.
        
        
        Returns
        -------
        self : `instance<cls>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's type is incorrect.
        """
        # application_id
        if application_id is ...:
            application_id = 0
        else:
            application_id = validate_application_id(application_id)
        
        # application_permissions
        if application_permissions is ...:
            application_permissions = Permission()
        else:
            application_permissions = validate_application_permissions(application_permissions)
        
        # attachment_size_limit
        if attachment_size_limit is ...:
            attachment_size_limit = LEVEL_0.attachment_size_limit
        else:
            attachment_size_limit = validate_attachment_size_limit(attachment_size_limit)
        
        # authorizer_user_ids
        if authorizer_user_ids is ...:
            authorizer_user_ids = None
        else:
            authorizer_user_ids = validate_authorizer_user_ids(authorizer_user_ids)
        
        # channel
        if channel is ...:
            channel = create_partial_channel_from_id(0, ChannelType.unknown, 0)
        else:
            channel = validate_channel(channel)
        
        # entitlements
        if entitlements is ...:
            entitlements = None
        else:
            entitlements = validate_entitlements(entitlements)
        
        # guild
        if guild is ...:
            guild = None
        else:
            guild = validate_guild(guild)
        
        # interaction & interaction_type
        if interaction_type is ...:
            interaction_type = InteractionType.none
        else:
            interaction_type = validate_type(interaction_type)
        
        if interaction is not ...:
            interaction = validate_interaction(interaction, interaction_type)
            warn(
                (
                    f'`{cls.__name__}.__new__`\'s `interaction` parameter is deprecated and scheduled for removal in '
                    f'2026 February. Please pass each parameter you would pass when creating the interaction as '
                    f'`**keyword_parameters` here.'
                ),
                FutureWarning,
                    stacklevel = 2,
            )
            _unpack_interaction_metadata_into_keyword_parameters(interaction_type, interaction, keyword_parameters)
        
        # message
        if message is ...:
            message = None
        else:
            message = validate_message(message)
        
        # resolved
        if resolved is ...:
            resolved = None
        else:
            resolved = validate_resolved(resolved)
        
        # token
        if token is ...:
            token = ''
        else:
            token = validate_token(token)
        
        # user
        if user is ...:
            user = ZEROUSER
        else:
            user = validate_user(user)
        
        # user_locale
        if user_locale is ...:
            user_locale = LOCALE_DEFAULT
        else:
            user_locale = validate_user_locale(user_locale)
        
        # user_permissions
        if user_permissions is ...:
            user_permissions = Permission()
        else:
            user_permissions = validate_user_permissions(user_permissions)
        
        # metadata
        metadata = interaction_type.metadata_type.from_keyword_parameters(keyword_parameters)
        if keyword_parameters:
            raise TypeError(
                f'Extra or unused keyword parameters: {keyword_parameters!r}.'
            )
        
        # Construct
        self = object.__new__(cls)
        self._async_task = None
        self._response_flags = RESPONSE_FLAG_NONE
        self.id = 0
        self.application_id = application_id
        self.application_permissions = application_permissions
        self.attachment_size_limit = attachment_size_limit
        self.authorizer_user_ids = authorizer_user_ids
        self.channel = channel
        self.entitlements = entitlements
        self.guild = guild
        self.message = message
        self.metadata = metadata
        self.resolved = resolved
        self.token = token
        self.type = interaction_type
        self.user = user
        self.user_locale = user_locale
        self.user_permissions = user_permissions
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new ``InteractionEvent`` with the given parameters.
        
        Parameters
        ----------
        data : `dict<str, object>`
            `INTERACTION_CREATE` dispatch event data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        # Need guild_id early, so we can create a guild if required.
        guild = parse_guild(data)
        
        # interaction_id
        interaction_id = parse_id(data)
        application_id = parse_application_id(data)
        application_permissions = parse_application_permissions(data)
        attachment_size_limit = parse_attachment_size_limit(data)
        authorizer_user_ids = parse_authorizer_user_ids(data)
        channel = parse_channel(data)
        entitlements = parse_entitlements(data)
        interaction_type = parse_type(data)
        metadata = interaction_type.metadata_type.from_data(data, 0 if guild is None else guild.id)
        user_locale = parse_user_locale(data)
        message = parse_message(data)
        resolved = parse_resolved(data, 0 if guild is None else guild.id)
        token = parse_token(data)
        user = parse_user(data, 0 if guild is None else guild.id)
        user_permissions = parse_user_permissions(data)
        
        self = object.__new__(cls)
        self._async_task = None
        self._response_flags = RESPONSE_FLAG_NONE
        self.application_id = application_id
        self.application_permissions = application_permissions
        self.attachment_size_limit = attachment_size_limit
        self.authorizer_user_ids = authorizer_user_ids
        self.channel = channel
        self.entitlements = entitlements
        self.guild = guild
        self.id = interaction_id
        self.message = message
        self.metadata = metadata
        self.resolved = resolved
        self.token = token
        self.type = interaction_type
        self.user = user
        self.user_locale = user_locale
        self.user_permissions = user_permissions
        
        if message is None:
            self._add_response_waiter()
        
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Returns the interaction event's json serializable representation.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether fields with their default values should be included as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = self.metadata.to_data(defaults = defaults, guild_id = self.guild_id)
        put_application_id(self.application_id, data, defaults)
        put_application_permissions(self.application_permissions, data, defaults)
        put_attachment_size_limit(self.attachment_size_limit, data, defaults)
        put_authorizer_user_ids(self.authorizer_user_ids, data, defaults)
        put_channel(self.channel, data, defaults)
        put_entitlements(self.entitlements, data, defaults)
        put_guild(self.guild, data, defaults)
        put_id(self.id, data, defaults)
        put_message(self.message, data, defaults)
        put_resolved(self.resolved, data, defaults, guild_id = self.guild_id)
        put_token(self.token, data, defaults)
        put_type(self.type, data, defaults)
        put_user(self.user, data, defaults, guild_id = self.guild_id)
        put_user_locale(self.user_locale, data, defaults)
        put_user_permissions(self.user_permissions, data, defaults)
        return data
    
    
    @classmethod
    def _create_empty(cls, interaction_id):
        """
        Creates a new partial interaction event with it's attributes set as their default value.
        
        Parameters
        ----------
        interaction_id : `int`
            The interaction event's identifier.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self._async_task = None
        self._response_flags = RESPONSE_FLAG_NONE
        self.application_id = 0
        self.application_permissions = Permission()
        self.attachment_size_limit = LEVEL_0.attachment_size_limit
        self.authorizer_user_ids = None
        self.channel = create_partial_channel_from_id(0, ChannelType.unknown, 0)
        self.entitlements = None
        self.guild = None
        self.id = interaction_id
        self.metadata = DEFAULT_INTERACTION_METADATA
        self.message = None
        self.resolved = None
        self.token = ''
        self.type = InteractionType.none
        self.user = ZEROUSER
        self.user_locale = LOCALE_DEFAULT
        self.user_permissions = Permission()
        return self
    
    
    @classmethod
    def precreate(
        cls,
        interaction_id,
        *,
        interaction = ...,
        **keyword_parameters,
    ):
        """
        Creates an interaction event. Not like ``.__new__``, ``.precreate`` allows setting ``.id`` as well.
        
        Since interaction events are not cached, there os no major advantage of using this method.
        
        Parameters
        ----------
        interaction_id : `int`
            The interaction's identifier.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters defining which attribute and how should be set.
        
        Other Parameters
        ----------
        application : `int`, `str`, ``Application``, Optional (Keyword only)
            Alternative for `application_id`.
        
        application_command_id : ``None | int | ApplicationCommand``, Optional (Keyword only)
            The represented application command's identifier number.
        
        application_command_name : `None | str`, Optional (Keyword only)
            The represented application command's name.
        
        application_id : `int`, `str`, ``Application``, Optional (Keyword only)
            The interaction's application's identifier.
        
        application_permissions : ``Permission``, `int`, Optional (Keyword only)
            The permissions granted to the application in the guild.
        
        attachment_size_limit : `int`, Optional (Keyword only)
            The size limit in bytes for each attachment.
        
        authorizer_user_ids : `None | dict<ApplicationIntegrationType | int, int | ClientUserBase>` \
                , Optional (Keyword only)
            The users' identifier who authorized the integration.
        
        channel : ``Channel``, Optional (Keyword only)
            The channel from where the interaction was called.
        
        component : ``None | InteractionComponent``, Optional (Keyword only)
            The interacted component.
        
        components : ``None | iterable<InteractionComponent>``, Optional (Keyword only)
            Submitted component values of a form submit interaction.
        
        custom_id : `None | str`, Optional (Keyword only)
            Form interaction's custom identifier.
        
        entitlements : `None`, `iterable` of ``Entitlement``, Optional (Keyword only)
            The applicable entitlements for the event's context.
        
        guild : ``None | Guild``, Optional (Keyword only)
            The guild from where the interaction was called from.
        
        interaction_type : ``InteractionType``, `int`, Optional (Keyword only)
            The interaction's type.
        
        message : ``None | Message``, Optional (Keyword only)
            The message from where the interaction was received.
        
        options : ``None | iterable<InteractionOption>``, Optional (Keyword only)
            Application command option representations. Like sub-command or parameter.
        
        resolved : ``None | Resolved``, Optional (Keyword only)
            Contains the received entities.
        
        target_id : `int`, Optional (Keyword only)
            The interaction's target's identifier. Applicable for context commands.
        
        target_type : ``None | int | ApplicationCommandTargetType``, Optional (Keyword only)
            The invoked application command's target type.
        
        token : `str`, Optional (Keyword only)
            Interaction's token used when responding on it.
        
        user : ``ClientUserBase``, Optional (Keyword only)
            The user who called the interaction.
        
        user_locale : ``Locale``, `str`, Optional (Keyword only)
            The selected language of the invoking user.
        
        user_permissions : ``Permission``, `int`, Optional (Keyword only)
            The user's permissions in the respective channel.
        
        Returns
        -------
        self : `instance<cls>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's type is incorrect.
        """
        interaction_id = validate_id(interaction_id)
        
        if keyword_parameters:
            processed = []
            
            # interaction & interaction_type
            try:
                interaction_type = keyword_parameters.pop('interaction_type')
            except KeyError:
                interaction_type = InteractionType.none
            else:
                interaction_type = validate_type(interaction_type)
                processed.append(('type', interaction_type))
            
            if interaction is not ...:
                interaction = validate_interaction(interaction, interaction_type)
                warn(
                    (
                        f'`{cls.__name__}.precreate`\'s `interaction` parameter is deprecated and scheduled for '
                        f'removal in 2026 February. Please pass each parameter you would pass when creating the '
                        f'interaction as `**keyword_parameters` here.'
                    ),
                    FutureWarning,
                    stacklevel = 2,
                )
                _unpack_interaction_metadata_into_keyword_parameters(interaction_type, interaction, keyword_parameters)
            
            metadata = interaction_type.metadata_type.from_keyword_parameters(keyword_parameters)
            processed.append(('metadata', metadata))
            
            extra = process_precreate_parameters(keyword_parameters, PRECREATE_FIELDS, processed)
            raise_extra(extra)
        
        else:
            processed = None
        
        self = cls._create_empty(interaction_id)
        
        if (processed is not None):
            for item in processed:
                setattr(self, *item)
        
        return self
    
    
    def copy(self):
        """
        Copies the interaction event.
        The returned interaction event is partial lacking internal fields.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new._async_task = None
        new._response_flags = RESPONSE_FLAG_NONE
        new.id = 0
        new.application_id = self.application_id
        new.application_permissions = self.application_permissions
        new.attachment_size_limit = self.attachment_size_limit
        
        authorizer_user_ids = self.authorizer_user_ids
        if (authorizer_user_ids is not None):
            authorizer_user_ids = authorizer_user_ids.copy()
        new.authorizer_user_ids = authorizer_user_ids
        
        new.type = self.type
        new.channel = self.channel
        
        entitlements = self.entitlements
        if (entitlements is not None):
            entitlements = (*entitlements,)
        new.entitlements = entitlements
        
        new.guild = self.guild
        
        new.message = self.message
        
        new.metadata = self.metadata.copy()
        
        resolved = self.resolved
        if (resolved is not None):
            resolved = resolved.copy()
        new.resolved = resolved
        
        new.token = self.token
        new.user = self.user
        new.user_locale = self.user_locale
        new.user_permissions = self.user_permissions
        return new
    
    
    def copy_with(
        self,
        *,
        application_id = ...,
        application_permissions = ...,
        attachment_size_limit = ...,
        authorizer_user_ids = ...,
        channel = ...,
        entitlements = ...,
        guild = ...,
        interaction = ...,
        interaction_type = ...,
        message = ...,
        resolved = ...,
        token = ..., 
        user = ...,
        user_locale = ...,
        user_permissions = ...,
        **keyword_parameters,
    ):
        """
        Copies the interaction event modifying it's defined fields.
        The returned interaction event is partial lacking internal fields.
        
        Parameters
        ----------
        application_id : `int`, `str`, Optional (Keyword only)
            The interaction's application's identifier.
        
        application_permissions : ``Permission``, `int`, Optional (Keyword only)
            The permissions granted to the application in the guild.
        
        attachment_size_limit : `int`, Optional (Keyword only)
            The size limit in bytes for each attachment.
        
        authorizer_user_ids : `None | dict<ApplicationIntegrationType | int, int | ClientUserBase>` \
                , Optional (Keyword only)
            The users' identifier who authorized the integration.
        
        channel : ``Channel``, Optional (Keyword only)
            The channel from where the interaction was called.
        
        channel_id : `int`, `str`, ``Channel``, Optional (Keyword only)
            The channel's identifier from where the interaction was called.
        
        entitlements : `None`, `iterable` of ``Entitlement``, Optional (Keyword only)
            The applicable entitlements for the event's context.
        
        guild : ``None | Guild```, Optional (Keyword only)
            The guild from where the interaction was called from.
        
        message : ``None | Message``, Optional (Keyword only)
            The message from where the interaction was received.
        
        resolved : ``None | Resolved``, Optional (Keyword only)
            Contains the received entities.
        
        token : `str`, Optional (Keyword only)
            Interaction's token used when responding on it.
        
        type : ``InteractionType``, `int`, Optional (Keyword only)
            The interaction's type.
        
        user : ``ClientUserBase``, Optional (Keyword only)
            The user who called the interaction.
        
        user_locale : ``Locale``, `str`, Optional (Keyword only)
            The selected language of the invoking user.
        
        user_permissions : ``Permission``, `int`, Optional (Keyword only)
            The user's permissions in the respective channel.
        
        **keyword_parameters : Keyword parameters
            Keyword parameters defining the type specific fields of the interaction.
        
        Other parameters
        ----------------
        application_command_id : ``None | int | ApplicationCommand``, Optional (Keyword only)
            The represented application command's identifier number.
        
        application_command_name : `None | str`, Optional (Keyword only)
            The represented application command's name.
        
        component : ``None | InteractionComponent``, Optional (Keyword only)
            The interacted component.
        
        components : ``None | iterable<InteractionComponent>``, Optional (Keyword only)
            Submitted component values of a form submit interaction.
        
        custom_id : `None | str`, Optional (Keyword only)
            Form interaction's custom identifier.
        
        options : ``None | iterable<InteractionOption>``, Optional (Keyword only)
            Application command option representations. Like sub-command or parameter.
        
        target_id : `int`, Optional (Keyword only)
            The interaction's target's identifier. Applicable for context commands.
        
        target_type : ``None | int | ApplicationCommandTargetType``, Optional (Keyword only)
            The invoked application command's target type.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's type is incorrect.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        # application_id
        if application_id is ...:
            application_id = self.application_id
        else:
            application_id = validate_application_id(application_id)
        
        # application_permissions
        if application_permissions is ...:
            application_permissions = self.application_permissions
        else:
            application_permissions = validate_application_permissions(application_permissions)
        
        # attachment_size_limit
        if attachment_size_limit is ...:
            attachment_size_limit = self.attachment_size_limit
        else:
            attachment_size_limit = validate_attachment_size_limit(attachment_size_limit)
        
        # authorizer_user_ids
        if authorizer_user_ids is ...:
            authorizer_user_ids = self.authorizer_user_ids
            if (authorizer_user_ids is not None):
                authorizer_user_ids = authorizer_user_ids.copy()
        
        else:
            authorizer_user_ids = validate_authorizer_user_ids(authorizer_user_ids)
        
        # guild
        if guild is ...:
            guild = self.guild
        else:
            guild = validate_guild(guild)
        
        # channel
        if channel is ...:
            channel = self.channel
        else:
            channel = validate_channel(channel)
        
        # entitlements
        if entitlements is ...:
            entitlements = self.entitlements
            if (entitlements is not None):
                entitlements = (*entitlements,)
        else:
            entitlements = validate_entitlements(entitlements)
        
        # interaction & interaction_type
        if interaction_type is ...:
            interaction_type = self.type
        else:
            interaction_type = validate_type(interaction_type)
        
        if interaction is not ...:
            interaction = validate_interaction(interaction, interaction_type)
            warn(
                (
                    f'`{type(self).__name__}.copy_with`\'s `interaction` parameter is deprecated and scheduled for '
                    f'removal in 2026 February. Please pass each parameter you would pass when creating the '
                    f'interaction as `**keyword_parameters` here.'
                ),
                FutureWarning,
                stacklevel = 2,
            )
            _unpack_interaction_metadata_into_keyword_parameters(interaction_type, interaction, keyword_parameters)
        
        # message
        if message is ...:
            message = self.message
        else:
            message = validate_message(message)
        
        # resolved
        if resolved is ...:
            resolved = self.resolved
            if (resolved is not None):
                resolved = resolved.copy()
        else:
            resolved = validate_resolved(resolved)
        
        # token
        if token is ...:
            token = self.token
        else:
            token = validate_token(token)
        
        # user
        if user is ...:
            user = self.user
        else:
            user = validate_user(user)
        
        # user_locale
        if user_locale is ...:
            user_locale = self.user_locale
        else:
            user_locale = validate_user_locale(user_locale)
        
        # user_permissions
        if user_permissions is ...:
            user_permissions = self.user_permissions
        else:
            user_permissions = validate_user_permissions(user_permissions)
        
        
        # metadata
        metadata = self.metadata
        metadata_type = interaction_type.metadata_type
        if metadata_type is type(metadata):
            metadata = metadata.copy_with_keyword_parameters(keyword_parameters)
        else:
            metadata = metadata_type.from_keyword_parameters(keyword_parameters)
        
        if keyword_parameters:
            raise TypeError(
                f'Extra or unused keyword parameters: {keyword_parameters!r}.'
            )
        
        # Construct
        new = object.__new__(type(self))
        new._async_task = None
        new._response_flags = RESPONSE_FLAG_NONE
        new.id = 0
        new.application_id = application_id
        new.application_permissions = application_permissions
        new.attachment_size_limit = attachment_size_limit
        new.authorizer_user_ids = authorizer_user_ids
        new.type = interaction_type
        new.channel = channel
        new.entitlements = entitlements
        new.guild = guild
        new.metadata = metadata
        new.message = message
        new.resolved = resolved
        new.token = token
        new.user = user
        new.user_locale = user_locale
        new.user_permissions = user_permissions
        
        return new


    async def wait_for_response_message(self, *, timeout = None):
        """
        Waits for response message. Applicable for application command interactions.
        
        This method is a coroutine.
        
        Parameters
        ----------
        timeout : `None`, `float` = `None`, Optional (Keyword only)
            The maximal time to wait for message before `TimeoutError` is raised.
        
        Returns
        -------
        message : ``Message``
            The received message.
        
        Raises
        ------
        TimeoutError
            Message was not received before timeout.
        """
        message = self.message
        if (message is not None):
            return message
        
        try:
            waiter = INTERACTION_EVENT_MESSAGE_WAITERS[self]
        except KeyError:
            waiter = Future(KOKORO)
            INTERACTION_EVENT_MESSAGE_WAITERS[self] = waiter
        
        waiter = shield(waiter, KOKORO)
        
        if (timeout is not None):
            waiter.apply_timeout(timeout)
        
        await waiter
        return self.message
    
    
    def __repr__(self):
        """Returns the representation of the event."""
        repr_parts = ['<', type(self).__name__]
        
        response_state_names = None
        response_state = self._response_flags
        
        if response_state == RESPONSE_FLAG_NONE:
            pass
        
        if response_state & RESPONSE_FLAG_DEFERRING:
            if response_state_names is None:
                response_state_names = []
            response_state_names.append('deferring')
        
        if response_state & RESPONSE_FLAG_DEFERRED:
            if response_state_names is None:
                response_state_names = []
            response_state_names.append('deferred')
        
        if response_state & RESPONSE_FLAG_RESPONDING:
            if response_state_names is None:
                response_state_names = []
            response_state_names.append('responding')
        
        if response_state & RESPONSE_FLAG_RESPONDED:
            if response_state_names is None:
                response_state_names = []
            response_state_names.append('responded')
        
        if response_state & RESPONSE_FLAG_EPHEMERAL:
            if response_state_names is None:
                response_state_names = []
            response_state_names.append('ephemeral')
        
        if (response_state_names is not None):
            repr_parts.append(' (')
            index = 0
            limit = len(response_state_names)
            while True:
                response_state_name = response_state_names[index]
                repr_parts.append(response_state_name)
                index += 1
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append('),')
        
        repr_parts.append(' type = ')
        metadata_type = self.type
        repr_parts.append(metadata_type.name)
        repr_parts.append(' ~ ')
        repr_parts.append(repr(metadata_type.value))
        
        authorizer_user_ids = self.authorizer_user_ids
        if (authorizer_user_ids is not None):
            repr_parts.append(', authorizer_user_ids = ')
            repr_parts.append(repr(authorizer_user_ids))
        
        guild = self.guild
        if guild is not None:
            repr_parts.append(', guild_id = ')
            repr_parts.append(repr(guild.id))
            
            repr_parts.append(', application_permissions = ')
            repr_parts.append(format(self.application_permissions, 'd'))
        
        # attachment_size_limit
        if guild is None:
            default_attachment_size_limit = LEVEL_0.attachment_size_limit
        else:
            default_attachment_size_limit = guild.attachment_size_limit
        
        attachment_size_limit = self.attachment_size_limit
        if attachment_size_limit != default_attachment_size_limit:
            repr_parts.append(', attachment_size_limit = ')
            repr_parts.append(repr(attachment_size_limit))
        
        
        repr_parts.append(', channel = ')
        repr_parts.append(repr(self.channel))
        
        # entitlements
        entitlements = self.entitlements
        if (entitlements is not None):
            length = len(entitlements)
            index = 0
            
            repr_parts.append(', entitlements = [')
            while True:
                entitlement = entitlements[index]
                index += 1
                
                repr_parts.append(repr(entitlement))
                
                if index == length:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']' )
        
        
        message = self.message
        if (message is not None):
            repr_parts.append(', message = ')
            repr_parts.append(repr(message))
        
        
        repr_parts.append(', user = ')
        repr_parts.append(repr(self.user))
        
        repr_parts.append(', user_locale = ')
        repr_parts.append(repr(self.user_locale.name))
        
        if guild is not None:
            repr_parts.append(', guild_locale = ')
            repr_parts.append(repr(guild.locale.name))
        
        repr_parts.append(', metadata = ')
        repr_parts.append(repr(self.metadata))
        
        # resolved
        resolved = self.resolved
        if (resolved is not None):
            repr_parts.append(', resolved = ')
            repr_parts.append(repr(resolved))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two interaction events are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # Shortcut
        if self is other:
            return True
        
        # id
        self_id = self.id
        other_id = other.id
        if self_id and other_id and (self.id != other.id):
            return False
        
        # application_id
        if self.application_id != other.application_id:
            return False
        
        # application_permissions
        if self.application_permissions != other.application_permissions:
            return False
        
        # attachment_size_limit
        if self.attachment_size_limit != other.attachment_size_limit:
            return False
        
        # authorizer_user_ids
        if self.authorizer_user_ids != other.authorizer_user_ids:
            return False
        
        # channel
        if self.channel is not other.channel:
            return False
        
        # entitlements
        if self.entitlements != other.entitlements:
            return False
        
        # guild
        if self.guild is not other.guild:
            return False
        
        # message
        if self.message is not other.message:
            return False
        
        # interaction
        if self.metadata != other.metadata:
            return False
        
        # resolved
        if self.resolved != other.resolved:
            return False
        
        # token
        if self.token != other.token:
            return False
        
        # type
        if self.type is not other.type:
            return False
        
        # user
        if self.user is not other.user:
            return False
        
        # user_locale
        if self.user_locale is not other.user_locale:
            return False
        
        # user_permissions
        if self.user_permissions != other.user_permissions:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the hash value of the interaction event."""
        self_id = self.id
        if self_id:
            return self_id
        
        hash_value = 0
        
        # application_id
        hash_value ^= self.application_id
        
        # application_permissions
        hash_value ^= self.application_permissions
        
        # attachment_size_limit
        guild = self.guild
        if guild is None:
            default_attachment_size_limit = LEVEL_0.attachment_size_limit
        else:
            default_attachment_size_limit = guild.attachment_size_limit
        
        attachment_size_limit = self.attachment_size_limit
        if attachment_size_limit != default_attachment_size_limit:
            hash_value ^= attachment_size_limit
        
        # authorizer_user_ids
        authorizer_user_ids = self.authorizer_user_ids
        if (authorizer_user_ids is not None):
            hash_value ^= len(authorizer_user_ids) << 8
            for integration_type, user_id in authorizer_user_ids.items():
                hash_value ^= (integration_type.value << 22) & user_id
        
        # channel
        hash_value ^= hash(self.channel)
        
        # entitlements
        entitlements = self.entitlements
        if (entitlements is not None):
            hash_value ^= hash(entitlements)
        
        # guild
        if (guild is not None):
            hash_value ^= hash(guild)
        
        # message
        message = self.message
        if (message is not None):
            hash_value ^= hash(self.message)
        
        # interaction
        hash_value ^= hash(self.metadata)
        
        # resolved
        resolved = self.resolved
        if (resolved is not None):
            hash_value ^= hash(resolved)
        
        # token
        hash_value ^= hash(self.token)
        
        # type
        hash_value ^= self.type.value
        
        # user
        hash_value ^= hash(self.user)
        
        # user_locale
        hash_value ^= hash(self.user_locale)
        
        # user_permissions
        hash_value ^= self.user_permissions
        
        return hash_value
    
    
    def is_unanswered(self):
        """
        Returns whether the event was not acknowledged and not acknowledging either.
        
        Returns
        -------
        is_unanswered : `bool`
        """
        return True if (self._response_flags == RESPONSE_FLAG_NONE) else False
    
    
    def is_acknowledging(self):
        """
        Returns whether the event is being acknowledged.
        
        Returns
        -------
        is_acknowledging : `bool`
        """
        return True if (self._response_flags & RESPONSE_FLAG_ACKNOWLEDGING) else False
    
    
    def is_acknowledged(self):
        """
        Returns whether the event is acknowledged.
        
        Returns
        -------
        is_acknowledged : `bool`
        """
        return True if (self._response_flags & RESPONSE_FLAG_ACKNOWLEDGED) else False
    
    
    def is_deferred(self):
        """
        Returns whether the event is deferred.
        
        Returns
        -------
        is_deferred : `bool`
        """
        response_state = self._response_flags
        if response_state & RESPONSE_FLAG_RESPONDED:
            return False
        
        if response_state & RESPONSE_FLAG_DEFERRED:
            return True
        
        return False
    
    
    def is_responding(self):
        """
        Returns whether the event it being responded.
        
        Returns
        -------
        is_responding : `bool`
        """
        return True if (self._response_flags & RESPONSE_FLAG_RESPONDING) else False
    
    
    def is_responded(self):
        """
        Returns whether was responded.
        
        Returns
        -------
        is_responded : `bool`
        """
        return True if (self._response_flags & RESPONSE_FLAG_RESPONDED) else False
    
    
    def is_expired(self):
        """
        Returns whether the event is already expired.
        
        Returns
        -------
        is_expired : `bool`
        """
        return now_as_id() > (self.id + INTERACTION_EVENT_EXPIRE_AFTER_ID_DIFFERENCE)
    
    
    def is_response_invoking_user_only(self):
        """
        Returns whether the event was responded as invoking user only.
        
        Returns
        -------
        is_response_invoking_user_only : `bool`
        """
        return True if (self._response_flags & RESPONSE_FLAG_EPHEMERAL) else False
    
    
    def __len__(self):
        """Helper for unpacking if needed."""
        return 3
    
    
    def __iter__(self):
        """
        Unpacks the event.
        
        This method is a generator.
        """
        yield self.type
        yield self.user
        yield self.metadata
    
    
    @property
    def channel_id(self):
        """
        Returns the interaction's channel's identifier.
        
        Returns
        -------
        channel_id : `int`
        """
        return self.channel.id
    
    
    @property
    def user_id(self):
        """
        Returns the interaction's user's identifier.
        
        Returns
        -------
        user_id : `int`
        """
        return self.user.id
    
    
    @property
    def guild_id(self):
        """
        Returns the interaction's guild's identifier.
        
        Returns
        -------
        guild_id : ``int`
        """
        guild = self.guild
        if guild is None:
            guild_id = 0
        else:
            guild_id = guild.id
        
        return guild_id
    
    
    @property
    def guild_locale(self):
        """
        Returns the interaction's guild's locale.
        
        Returns
        -------
        guild_locale : ``Locale``
        """
        guild = self.guild
        if guild is None:
            guild_locale = LOCALE_DEFAULT
        else:
            guild_locale = guild.locale
        
        return guild_locale
    
    
    @property
    def client(self):
        """
        Returns the interaction's client.
        
        Returns
        -------
        client : `Client`
        
        Raises
        ------
        RuntimeError
            Client could not be identified.
        """
        try:
            return APPLICATION_ID_TO_CLIENT[self.application_id]
        except KeyError:
            raise RuntimeError(
                f'Client of {self!r} could not be identified.'
            ) from None
    
    
    @property
    def voice_client(self):
        """
        Returns the voice client of the interaction's client in it's guild.
        
        Returns
        -------
        voice_client : `None`, ``VoiceClient``
        """
        try:
            client = APPLICATION_ID_TO_CLIENT[self.application_id]
        except KeyError:
            voice_client = None
        else:
            guild = self.guild
            if guild is None:
                voice_client = None
            else:
                voice_client = client.voice_clients.get(guild.id, None)
        
        return voice_client
    
    
    def has_entitlement(self, entitlement):
        """
        Returns whether the interaction event hast he given entitlement.
        
        Parameters
        ----------
        entitlement : ``Entitlement``
            The entitlement to look for.
        
        Returns
        -------
        has_entitlement : `bool`
        """
        entitlements = self.entitlements
        if entitlements is None:
            return False
        
        return entitlement in entitlements
    
    
    def iter_entitlements(self):
        """
        Iterates over the entitlements of the interaction event.
        
        This method is an iterable generator.
        
        Yields
        ------
        entitlement : ``Entitlement``
        """
        entitlements = self.entitlements
        if (entitlements is not None):
            yield from entitlements
    
    
    def has_sku(self, sku):
        """
        Returns whether any of the interaction event's entitlements grants access to the given stock keeping unit.
        
        Parameters
        ----------
        sku : ``SKU``
            Stock keeping unit to look for.
        
        Returns
        -------
        has_sku : `bool`
        """
        entitlements = self.entitlements
        if (entitlements is None):
            return False
        
        sku_id = sku.id
        
        for entitlement in entitlements:
            if entitlement.sku_id == sku_id:
                return True
        
        return False
    
    
    def _add_response_waiter(self):
        """
        Adds the interaction event to response waiters.
        
        Called when the interaction is application command one, to resolve it's ``.message`` attribute when created.
        """
        INTERACTION_EVENT_RESPONSE_WAITERS[self.id] = self
    
    
    async def _wait_for_async_task_completion(self):
        """
        Waits for async task's completion.
        
        This method is a coroutine.
        """
        async_task = self._async_task
        if (async_task is not None) and (async_task is not KOKORO.current_task):
            try:
                await async_task
            finally:
                self._async_task = None
    
    
    # Field Proxies
    
    @property
    @copy_docs(InteractionMetadataBase.application_command_id)
    def application_command_id(self):
        return self.metadata.application_command_id
    
    @application_command_id.setter
    def application_command_id(self, value):
        self.metadata.application_command_id = validate_application_command_id(value)
    
    
    @property
    @copy_docs(InteractionMetadataBase.application_command_name)
    def application_command_name(self):
        return self.metadata.application_command_name
    
    @application_command_name.setter
    def application_command_name(self, value):
        self.metadata.application_command_name = validate_application_command_name(value)
    
    
    @property
    @copy_docs(InteractionMetadataBase.components)
    def components(self):
        return self.metadata.components
    
    @components.setter
    def components(self, value):
        self.metadata.components = validate_components(value)
    
    
    @property
    @copy_docs(InteractionMetadataBase.component)
    def component(self):
        return self.metadata.component
    
    @component.setter
    def component(self, value):
        self.metadata.component = validate_component(value)
    
    
    @property
    @copy_docs(InteractionMetadataBase.custom_id)
    def custom_id(self):
        return self.metadata.custom_id
    
    @custom_id.setter
    def custom_id(self, value):
        self.metadata.custom_id = validate_custom_id(value)
    
    
    @property
    @copy_docs(InteractionMetadataBase.options)
    def options(self):
        return self.metadata.options
    
    @options.setter
    def options(self, value):
        self.metadata.options = validate_options(value)
    
    
    @property
    @copy_docs(InteractionMetadataBase.target_id)
    def target_id(self):
        return self.metadata.target_id
    
    @target_id.setter
    def target_id(self, value):
        self.metadata.target_id = validate_target_id(value)
    
    
    @property
    @copy_docs(InteractionMetadataBase.target_type)
    def target_type(self):
        return self.metadata.target_type
    
    @target_type.setter
    def target_type(self, value):
        self.metadata.target_type = validate_target_type(value)
    
    # application command
    
    @property
    def target(self):
        """
        Returns the interaction's target. Applicable if the interaction was invoked by a context command.
        
        Returns
        -------
        entity : ``None | Attachment | Channel | ClientUserBase | Role | Message``
        """
        target_id = self.target_id
        if target_id:
            return self.resolve_entity(target_id)
    
    
    # application command autocomplete
    
    @copy_docs(InteractionMetadataBase.iter_options)
    def iter_options(self):
        return (yield from self.metadata.iter_options())
    
    
    @property
    @copy_docs(InteractionMetadataBase.focused_option)
    def focused_option(self):
        return self.metadata.focused_option
    
    
    @copy_docs(InteractionMetadataBase.get_non_focused_values)
    def get_non_focused_values(self):
        return self.metadata.get_non_focused_values()
    
    
    @copy_docs(InteractionMetadataBase.get_value_of)
    def get_value_of(self, *option_names):
        return self.metadata.get_value_of(*option_names)
    
    # Message component
    
    def iter_entities(self):
        """
        Iterates over the entities that were selected by the user of a select component interaction.
        
        This method is an iterable generator.
        
        Yields
        ------
        entity : ``str | Channel | ClientUserbase | Role``
        """
        resolved = self.resolved
        for custom_id, component_type, value_or_values in self.iter_custom_ids_and_values():
            
            iter_resolve = component_type.iter_resolve
            if (iter_resolve is None):
                continue
            
            yield from iter_resolve(resolved, value_or_values)
    
    
    @property
    def entities(self):
        """
        Returns the entities that were selected by the user of a select component interaction.
        
        Returns
        -------
        entities : ``list<Channel | ClientUserbase | Role>``
        """
        return [*self.iter_entities()]
    
    
    # form submit
    
    @copy_docs(InteractionMetadataBase.iter_components)
    def iter_components(self):
        return (yield from self.metadata.iter_components())
    
    
    @copy_docs(InteractionMetadataBase.iter_custom_ids_and_values)
    def iter_custom_ids_and_values(self):
        return (yield from self.metadata.iter_custom_ids_and_values())
    
    
    def get_custom_id_value_relation(self):
        """
        Returns a dictionary with `custom_id` to `value` relation.
        
        Returns
        -------
        custom_id_is_value_values_relation : `dict<str, (ComponentType, None | str | tuple<str>)>`
        """
        custom_id_value_relation = {}
        
        for custom_id, component_type, value_or_values in self.iter_custom_ids_and_values():
            if (value_or_values is not None):
                custom_id_value_relation[custom_id] = (component_type, value_or_values)
        
        return custom_id_value_relation
    
    
    def get_value_for(self, custom_id_to_match):
        """
        Returns the value for the given `custom_id`.
        
        Parameters
        ----------
        custom_id_to_match : `str`
            A respective components `custom_id` to match.
        
        Returns
        -------
        is_value_values : `(ComponentType, None | str | tuple<str>)` 
        """
        for custom_id, component_type, value_or_values in self.iter_custom_ids_and_values():
            if (custom_id == custom_id_to_match):
                return component_type, value_or_values
        
        return ComponentType.none, None
    
    
    def get_match_and_value(self, matcher):
        """
        Gets a `custom_id`'s value matching the given `matcher`.
        
        Parameters
        ----------
        matcher : `callable`
            Matcher to call on a `custom_id`.
            
            Should accept the following parameters:
            
            +-----------+-----------+
            | Name      | Type      |
            +===========+===========+
            | custom_id | `str`     |
            +-----------+-----------+
            
            Should return non-`None` on success.
        
        Returns
        -------
        match_and_is_value_values : `(None | object, ComponentType, None | str | tuple<str>)`
        """
        for custom_id, component_type, value_or_values in self.iter_custom_ids_and_values():
            match = matcher(custom_id)
            if (match is not None):
                return match, component_type, value_or_values
        
        return None, ComponentType.none, None
    
    
    def iter_matches_and_values(self, matcher):
        """
        Gets a `custom_id`'s value matching the given `matcher`.
        
        This method is an iterable generator.
        
        Parameters
        ----------
        matcher : `callable`
            Matcher to call on a `custom_id`
            
            Should accept the following parameters:
            
            +-----------+-----------+
            | Name      | Type      |
            +===========+===========+
            | custom_id | `str`     |
            +-----------+-----------+
            
            Should return non-`None` on success.
        
        Yields
        -------
        match_and_is_value_values : `(object, ComponentType, None | str | tuple<str>)`
        """
        for custom_id, component_type, value_or_values in self.iter_custom_ids_and_values():
            match = matcher(custom_id)
            if (match is not None):
                yield match, component_type, value_or_values
    
    # resolved
    
    @copy_docs(Resolved.resolve_attachment)
    def resolve_attachment(self, attachment_id):
        resolved = self.resolved
        if (resolved is not None):
            return resolved.resolve_attachment(attachment_id)
    
    
    @copy_docs(Resolved.resolve_channel)
    def resolve_channel(self, channel_id):
        resolved = self.resolved
        if (resolved is not None):
            return resolved.resolve_channel(channel_id)
    
    
    @copy_docs(Resolved.resolve_role)
    def resolve_role(self, role_id):
        resolved = self.resolved
        if (resolved is not None):
            return resolved.resolve_role(role_id)
    
    
    @copy_docs(Resolved.resolve_message)
    def resolve_message(self, message_id):
        resolved = self.resolved
        if (resolved is not None):
            return resolved.resolve_message(message_id)
    
    
    @copy_docs(Resolved.resolve_user)
    def resolve_user(self, user_id):
        resolved = self.resolved
        if (resolved is not None):
            return resolved.resolve_user(user_id)
    
    
    @copy_docs(Resolved.resolve_mentionable)
    def resolve_mentionable(self, mentionable_id):
        resolved = self.resolved
        if (resolved is not None):
            return resolved.resolve_mentionable(mentionable_id)
    
    
    @copy_docs(Resolved.resolve_entity)
    def resolve_entity(self, entity_id):
        resolved = self.resolved
        if (resolved is not None):
            return resolved.resolve_entity(entity_id)
    
    
    # authorized_user_ids
    
    def get_authorizer_user_id(self, integration_type):
        """
        Gets the authorizer user's identifier for the given integration type.
        
        Parameters
        ----------
        integration_type : `ApplicationIntegrationType | int`
            Integration type to query for.
        
        Returns
        -------
        user_id : `int`
        """
        authorizer_user_ids = self.authorizer_user_ids
        if (authorizer_user_ids is not None):
            try:
                return authorizer_user_ids[integration_type]
            except KeyError:
                pass
        
        return 0
    
    
    def get_authorizer_user(self, integration_type):
        """
        Gets the authorizer user for the given integration type.
        
        Parameters
        ----------
        integration_type : `ApplicationIntegrationType | int`
            Integration type to query for.
        
        Returns
        -------
        user : `None | ClientUserBase`
        """
        user_id = self.get_authorizer_user_id(integration_type)
        if user_id:
            return create_partial_user_from_id(user_id)
    
    
    # Deprecations
    
    
    @property
    def values(self):
        """
        Deprecated and scheduled for removal in 2026 February. Please use ``.iter_custom_ids_and_values`` instead.
        """
        warn(
            (
                f'`{type(self).__name__}.values` is deprecated and is scheduled for removal in 2026 February. '
                f'Please use `.iter_custom_ids_and_values` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        for custom_id, component_type, values_or_values in self.iter_custom_ids_and_values():
            if values_or_values is None:
                return
            
            if component_type.layout_flags.holds_value_multiple:
                return values_or_values
            
            return (values_or_values,)
    
    
    def iter_values(self):
        """
        Deprecated and scheduled for removal in 2026 February. Please use ``.iter_custom_ids_and_values`` instead.
        """
        warn(
            (
                f'`{type(self).__name__}.iter_values` is deprecated and is scheduled for removal in 2026 February. '
                f'Please use `.iter_custom_ids_and_values` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        for custom_id, component_type, values_or_values in self.iter_custom_ids_and_values():
            if (values_or_values is None):
                return
            
            if component_type.layout_flags.holds_value_multiple:
                yield from values_or_values
                return
            
            yield values_or_values
            return
    
    
    @property
    def value(self):
        """
        Deprecated and scheduled for removal in 2026 February. Please use ``.iter_custom_ids_and_values`` instead.
        """
        warn(
            (
                f'`{type(self).__name__}.value` is deprecated and is scheduled for removal in 2026 February. '
                f'Please use `.iter_custom_ids_and_values` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        for custom_id, component_type, values_or_values in self.iter_custom_ids_and_values():
            if (values_or_values is None):
                return
            
            if component_type.layout_flags.holds_value_multiple:
                return values_or_values[0]
            
            return values_or_values
    
    
    @property
    def component_type(self):
        """
        Deprecated and scheduled for removal in 2026 February. Please use ``.iter_custom_ids_and_values`` instead.
        """
        warn(
            (
                f'`{type(self).__name__}.component_type` is deprecated and is scheduled for removal in 2026 February. '
                f'Please use `.iter_custom_ids_and_values` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        for custom_id, component_type, values_or_values in self.iter_custom_ids_and_values():
            return component_type
        
        return ComponentType.none
    
    
    @property
    def interaction(self):
        """
        Deprecated and scheduled for removal in 2026 February. Please use ``.metadata`` instead.
        """
        warn(
            (
                f'`{type(self).__name__}.interaction` is deprecated and is scheduled for removal in 2026 February. '
                f'Please use `.metadata` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        return self.metadata



__all__ = ()

from ...color import Color
from ...field_parsers import int_parser_factory, nullable_functional_array_parser_factory, preinstanced_parser_factory
from ...field_putters import int_putter_factory, nullable_functional_array_optional_putter_factory, preinstanced_putter_factory
from ...field_validators import int_conditional_validator_factory, nullable_object_array_validator_factory, preinstanced_validator_factory

from .constants import GRADIENT_ANGLE_MAX, GRADIENT_ANGLE_MIN, INTENSITY_MAX, INTENSITY_MIN
from .preinstanced import SharedClientThemeBaseTheme


# base_theme

parse_base_theme = preinstanced_parser_factory(
    'base_theme', SharedClientThemeBaseTheme, SharedClientThemeBaseTheme.none
)
put_base_theme = preinstanced_putter_factory('base_theme')
validate_base_theme = preinstanced_validator_factory('base_theme', SharedClientThemeBaseTheme)


# colors

parse_colors = nullable_functional_array_parser_factory('colors', lambda value : Color(value, 16))
put_colors = nullable_functional_array_optional_putter_factory('colors', lambda color : format(color, 'X'))
validate_colors = nullable_object_array_validator_factory('colors', Color)


# gradient_angle

parse_gradient_angle = int_parser_factory('gradient_angle', GRADIENT_ANGLE_MIN)
put_gradient_angle = int_putter_factory('gradient_angle')
validate_gradient_angle = int_conditional_validator_factory(
    'gradient_angle',
    GRADIENT_ANGLE_MIN,
    (lambda gradient_angle : gradient_angle >= GRADIENT_ANGLE_MIN and gradient_angle <= GRADIENT_ANGLE_MAX),
    f'>= {GRADIENT_ANGLE_MIN} and <= {GRADIENT_ANGLE_MAX}',
)


# intensity

parse_intensity = int_parser_factory('base_mix', INTENSITY_MIN)
put_intensity = int_putter_factory('base_mix')
validate_intensity = int_conditional_validator_factory(
    'intensity',
    INTENSITY_MIN,
    (lambda intensity : intensity >= INTENSITY_MIN and intensity <= INTENSITY_MAX),
    f'>= {INTENSITY_MIN} and <= {INTENSITY_MAX}',
)



__all__ = ('command_sort_key', 'normalize_command_name',)


def normalize_command_name(command_name):
    """
    Normalises the given command name.
    
    Parameters
    ----------
    command_name : `str`
        The command name to normalize.
    
    Returns
    -------
    command_name : `str`
        The command's name.
    """
    return '-'.join(command_name.lower().replace('_', ' ').replace('-', ' ').split())


def get_function_name(function, name):
    """
    Tries to get the given function's name.
    
    Parameters
    ----------
    function : `object`
        The function to get it's name.
    name : `None`, `str`
        Alternative name to use if given.
    
    Returns
    -------
    name : `str`
    """
    if name is None:
        try:
            name = function.__name__
        except AttributeError:
            name = type(function).__name__
    
    return normalize_command_name(name)


def normalize_command_description(description):
    """
    Normalizes the given description.
    
    Parameters
    ----------
    description : `None`, `str`
        The description to normalise.
    
    Returns
    -------
    description : `None`, `str`
    """
    if description is None:
        return description
    
    lines = description.split('\n')
    if not lines:
        return None
    
    for index in range(len(lines)):
        lines[index] = lines[index].rstrip()
    
    indentation_to_remove = 0
    
    while True:
        for line in lines:
            if len(line) <= indentation_to_remove:
                continue
            
            if line[indentation_to_remove] in (' ', '\t'):
                continue
            
            break
        
        else:
            indentation_to_remove += 1
            continue
        
        break
    
    
    for index in range(len(lines)):
        lines[index] = lines[index][indentation_to_remove:]
    
    while True:
        if lines[-1]:
            break
        
        del lines[-1]
        if not lines:
            return None
    
    while True:
        if lines[0]:
            break
        
        del lines[0]
        continue
    
    # Add postprocessing modifications / line if required
    documentation_parts = []
    index = 0
    line_count = len(lines)
    
    while True:
        documentation_parts.append(lines[index])
        
        index += 1
        if index == line_count:
            break
        
        documentation_parts.append('\n')
        continue
    
    return ''.join(documentation_parts)


def get_function_description(function, description):
    """
    Tries to get the given function's description.
    
    Parameters
    ----------
    function : `object`
        The function to get it's description.
    description : `None`, `str`
        Alternative description to use if given.
    
    Returns
    -------
    description : `None`, `str`
    """
    if description is None:
        try:
            description = function.__doc__
        except AttributeError:
            description = None
    
    return normalize_command_description(description)


def normalize_aliases(aliases, name):
    """
    Tries to normalize the given
    
    Parameters
    ----------
    aliases : `None | str | iterable<str>`
        Alternative names for the command.
    name : `str`
        The command's name.
    
    Returns
    -------
    normalized_aliases : `None`, `str` of `str`
    """
    if aliases is None:
        return None
    
    
    if isinstance(aliases, str):
        alter = normalize_command_name(aliases)
        if alter == name:
            return None
        
        return {alter}
    
    
    iterator = getattr(aliases, '__iter__', None)
    if iterator is None:
        raise TypeError(
            f'`aliases` can be `None | str | iterable<str>`, got {type(aliases).__name__}; {aliases!r}.'
        )
    
    normalized_aliases = None
    
    for alter in aliases:
        if not isinstance(alter, str):
            raise TypeError(
                f'`aliases` elements can be `str`, got {alter.__class__.__name__}; {alter!r}; aliases={aliases!r}.'
            )
        
        alter = normalize_command_name(aliases)
        if alter == name:
            continue
        
        if (normalized_aliases is None):
            normalized_aliases = set()
        
        normalized_aliases.add(alter)
    
    
    return normalized_aliases


def command_sort_key(command):
    """
    Sort key used to sort commands by name.
    
    Parameters
    ----------
    command : ``Command``, ``CommandCategory``, ``CommandFunction``
        Command to get it's sort key of.
    
    Returns
    -------
    sort_key : `str`
    """
    name = command.name
    if name is None:
        name = ''
    
    return name



__all__ = ('EmbedFieldBase',)

from scarletio import RichAttributeErrorBaseType


class EmbedFieldBase(RichAttributeErrorBaseType):
    """
    Base type for embed fields.
    """
    __slots__ = ()
    
    def __new__(cls):
        """
        Creates a new embed field with the given parameters.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        return object.__new__(cls)
    
    
    def __len__(self):
        """Returns the embed field's contents' length."""
        return 0
    
    
    def __bool__(self):
        """Returns whether the embed field is not empty."""
        return False
    
    
    def __repr__(self):
        """Returns the embed field's representation."""
        repr_parts = ['<', self.__class__.__name__]
        self._put_repr_parts_into(repr_parts)
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def _put_repr_parts_into(self, repr_parts):
        """
        Appends the embed field's representation parts.
        
        Parameters
        ----------
        repr_parts : `list<str>`
        """
        pass
    
    
    def __hash__(self):
        """Returns the embed field's hash value."""
        return 0
    
    
    def __eq__(self, other):
        """Returns whether the two embed fields are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_equal_same_type(other)
    
    
    def _is_equal_same_type(self, other):
        """
        Returns whether the two embed fields.
        
        Helper method for ``.__eq__``
        
        Parameters
        ----------
        other : `instance<type<self>>`
            The other instance. Must be from the same type.
        
        Returns
        -------
        is_equal : `bool`
        """
        return True
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates an embed field from the data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Embed field data.
        
        Returns
        -------
        embed_field : `instance<cls>`
        """
        return object.__new__(cls)
    
    
    def to_data(self, *, defaults = False, include_internals = False):
        """
        Converts the embed field to json serializable object representing it.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default values should be included as well.
        
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether we want to include identifiers as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        return {}
    
    
    def clean_copy(self, guild = None):
        """
        Creates a clean copy of the embed field by removing the mentions in it's contents.
        
        Parameters
        ----------
        guild : ``None | Guild`` = `None`, Optional
            The respective guild as a context to look up guild specific names of entities.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        return object.__new__(type(self))
    
    
    def copy(self):
        """
        Copies the embed field returning a new one.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        return object.__new__(type(self))
    
    
    def copy_with(self):
        """
        Copies the embed field with the given parameters.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        return object.__new__(type(self))
    
    
    @property
    def contents(self):
        """
        Returns the contents of the embed field.
        
        Returns
        -------
        contents : `list<str>`
        """
        return [*self.iter_contents()]
    
    
    def iter_contents(self):
        """
        Iterates over the contents of the embed field.
        
        This method is an iterable generator.
        
        Yields
        ------
        content : `str`
        """
        return
        yield



__all__ = ('InteractionComponentMetadataButton',)

from scarletio import copy_docs, include

from .base import InteractionComponentMetadataBase
from .fields import parse_custom_id, put_custom_id, validate_custom_id


ComponentType = include('ComponentType')


class InteractionComponentMetadataButton(InteractionComponentMetadataBase):
    """
    Interaction component metadata representing a button component.
    
    Attributes
    ----------
    custom_id : `None | str`
        Custom identifier to detect which component was clicked (or used) by the user.
    """
    __slots__ = ('custom_id')
    
    def __new__(cls, *, custom_id = ...):
        """
        Creates a new interaction component metadata from the given fields.
        
        Parameters
        ----------
        custom_id : `None | str`, Optional (Keyword only)
            Custom identifier to detect which component was clicked (or used) by the user.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # custom_id
        if custom_id is ...:
            custom_id = None
        else:
            custom_id = validate_custom_id(custom_id)
        
        # Construct.
        self = object.__new__(cls)
        self.custom_id = custom_id
        return self
    
    
    @classmethod
    def from_keyword_parameters(cls, keyword_parameters):
        """
        Creates a new interaction component metadata from the given keyword parameters.
        
        Parameters
        ----------
        keyword_parameters : `dict<str, object>`
            Keyword parameters to build the metadata from.
        
        Returns
        -------
        self : `instance<type<cls>>`
        
        Raises
        ------
        TypeError
            - If a keyword parameter's type is incorrect.
        ValueError
            - If a keyword parameter's value is incorrect.
        """
        return cls(
            custom_id = keyword_parameters.pop('custom_id', ...),
        )
    
    
    @copy_docs(InteractionComponentMetadataBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', type(self).__name__]
        
        # custom_id
        repr_parts.append(' custom_id = ')
        repr_parts.append(repr(self.custom_id))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(InteractionComponentMetadataBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # custom_id
        custom_id = self.custom_id
        if (custom_id is not None):
            hash_value ^= hash(custom_id)
        
        return hash_value
    
    
    @copy_docs(InteractionComponentMetadataBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        # custom_id
        if self.custom_id != other.custom_id:
            return False
        
        return True
    
    
    @copy_docs(InteractionComponentMetadataBase._match_to_component)
    def _match_to_component(self, other):
        # custom_id
        if self.custom_id != other.custom_id:
            return False
        
        return True
    
    
    @classmethod
    @copy_docs(InteractionComponentMetadataBase.from_data)
    def from_data(cls, data):
        self = object.__new__(cls)
        self.custom_id = parse_custom_id(data)
        return self
    
    
    @copy_docs(InteractionComponentMetadataBase.to_data)
    def to_data(self, *, defaults = False):
        data = {}
        put_custom_id(self.custom_id, data, defaults)
        return data
    
    
    @copy_docs(InteractionComponentMetadataBase.copy)
    def copy(self):
        new = object.__new__(type(self))
        
        # custom_id
        new.custom_id = self.custom_id
        
        return new
    
    
    def copy_with(self, *, custom_id = ...):
        """
        Copies the interaction component metadata with the given fields.
        
        Parameters
        ----------
        custom_id : `None | str`, Optional (Keyword only)
            Custom identifier to detect which component was clicked (or used) by the user.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - Extra or unused parameters.
        ValueError
            - If a parameter's value is incorrect.
        """
        # custom_id
        if custom_id is ...:
            custom_id = self.custom_id
        else:
            custom_id = validate_custom_id(custom_id)
        
        # Construct.
        new = object.__new__(type(self))
        new.custom_id = custom_id
        return new
    
    
    @copy_docs(InteractionComponentMetadataBase.copy_with_keyword_parameters)
    def copy_with_keyword_parameters(self, keyword_parameters):
        return self.copy_with(
            custom_id = keyword_parameters.pop('custom_id', ...),
        )
    
    
    @copy_docs(InteractionComponentMetadataBase.iter_custom_ids_and_values)
    def iter_custom_ids_and_values(self):
        custom_id = self.custom_id
        if (custom_id is not None):
            yield (custom_id, ComponentType.button, None)



__all__ = ('ApplicationRoleConnectionMetadataType', 'ApplicationRoleConnectionValueType')

from ...bases import Preinstance as P, PreinstancedBase
from ...utils import datetime_to_timestamp, timestamp_to_datetime_soft


class ApplicationRoleConnectionValueType(PreinstancedBase, value_type = int):
    """
    Type information for application role connection metadata values'.
    
    > This type is wrapper side only.
    
    Attributes
    ----------
    deserializer : `FunctionType`
        Function used to deserialize values.
    
    name : `str`
        The name of the value type.
    
    serializer : `FunctionType`
        Function used to serialize values.
    
    value : `int`
        The identifier of the value type.
    
    Type Attributes
    ---------------
    Every predefined value type can be accessed as type attribute as well:
    
    +---------------------------+---------------------------+-------+
    | Type attribute name       | name                      | value |
    +===========================+===========================+=======+
    | none                      | none                      | 0     |
    +---------------------------+---------------------------+-------+
    | integer                   | integer                   | 1     |
    +---------------------------+---------------------------+-------+
    | datetime                  | datetime                  | 2     |
    +---------------------------+---------------------------+-------+
    | boolean                   | boolean                   | 3     |
    +---------------------------+---------------------------+-------+
    """
    __slots__ = ('deserializer', 'serializer')
    
    DESERIALIZER_DEFAULT = lambda value: value
    SERIALIZER_DEFAULT = DESERIALIZER_DEFAULT
    
    def __new__(cls, value, name = None, deserializer = None, serializer = None):
        """
        Creates an application role connection value type.
        
        Parameters
        ----------
        value : `int`
            The identifier value of the application role connection value type.
        
        name : `None | str` = `None`, Optional
            The default name of the application role connection value type.
        
        deserializer : `None | FunctionType` = `None`, Optional
            Function used to deserialize values.
        
        serializer : `None | FunctionType` = `None`, Optional
            Function used to serialize values.
        """
        if deserializer is None:
            deserializer = cls.DESERIALIZER_DEFAULT
        
        if serializer is None:
            serializer = cls.SERIALIZER_DEFAULT
        
        self = PreinstancedBase.__new__(cls, value, name)
        self.deserializer = deserializer
        self.serializer = serializer
        return self
    
    
    # predefined
    none = P(
        0,
        'none',
        DESERIALIZER_DEFAULT,
        SERIALIZER_DEFAULT,
    )
    integer = P(
        1,
        'integer',
        (lambda value: int(value) if value.isdecimal() else None),
        (lambda value: format(value, 'd')),
    )
    datetime = P(
        2,
        'datetime',
        timestamp_to_datetime_soft,
        datetime_to_timestamp,
    )
    boolean = P(
        3,
        'boolean',
        (lambda value: {'0': False, '1': True}.get(value, None)),
        (lambda value: '1' if value else '0'),
    )


class ApplicationRoleConnectionMetadataType(PreinstancedBase, value_type = int):
    """
    Represents an application role connection type.
    
    Attributes
    ----------
    name : `str`
        The name of the application role connection type.
    
    value : `int`
        The Discord side identifier value of the application role connection type.
    
    value_type : ``ApplicationRoleConnectionValueType``
        Additional information describing the metadata's value's type.
    
    Type Attributes
    ---------------
    Every predefined application role connection type can be accessed as type attribute as well:
    
    +---------------------------+---------------------------+-------+---------------+
    | Type attribute name       | name                      | value | value type    |
    +===========================+===========================+=======+===============+
    | none                      | none                      | 0     | none          |
    +---------------------------+---------------------------+-------+---------------+
    | integer_less_or_equal     | integer less or equal     | 1     | integer       |
    +---------------------------+---------------------------+-------+---------------+
    | integer_greater_or_equal  | integer greater or equal  | 2     | integer       |
    +---------------------------+---------------------------+-------+---------------+
    | integer_equal             | integer equal             | 3     | integer       |
    +---------------------------+---------------------------+-------+---------------+
    | integer_not_equal         | integer not equal         | 4     | integer       |
    +---------------------------+---------------------------+-------+---------------+
    | datetime_less_or_equal    | datetime less or equal    | 5     | datetime      |
    +---------------------------+---------------------------+-------+---------------+
    | datetime_greater_or_equal | datetime greater or equal | 6     | datetime      |
    +---------------------------+---------------------------+-------+---------------+
    | boolean_equal             | boolean equal             | 7     | boolean       |
    +---------------------------+---------------------------+-------+---------------+
    | boolean_not_equal         | boolean not equal         | 8     | boolean       |
    +---------------------------+---------------------------+-------+---------------+
    """
    __slots__ = ('value_type',)
    
    def __new__(cls, value, name = None, value_type = None):
        """
        Creates an application role connection metadata type.
        
        Parameters
        ----------
        value : `int`
            The Discord side identifier value of the application role connection metadata type.
        
        name : `str`
            The default name of the application role connection metadata type.
        
        value_type : `None | ApplicationRoleConnectionValueType` = `None`, Optional
            Additional information describing the metadata's value's type.
        """
        if value_type is None:
            value_type = ApplicationRoleConnectionValueType.none
        
        self = PreinstancedBase.__new__(cls, value, name)
        self.value_type = value_type
        return self
    
    
    # predefined
    none = P(0, 'none', ApplicationRoleConnectionValueType.none)
    integer_less_or_equal = P(1, 'integer less or equal', ApplicationRoleConnectionValueType.integer)
    integer_greater_or_equal = P(2, 'integer greater or equal', ApplicationRoleConnectionValueType.integer)
    integer_equal = P(3, 'integer equal', ApplicationRoleConnectionValueType.integer)
    integer_not_equal = P(4, 'integer not equal', ApplicationRoleConnectionValueType.integer)
    datetime_less_or_equal = P(5, 'datetime less or equal', ApplicationRoleConnectionValueType.datetime)
    datetime_greater_or_equal = P(6, 'datetime not equal', ApplicationRoleConnectionValueType.datetime)
    boolean_equal = P(7, 'boolean equal', ApplicationRoleConnectionValueType.boolean)
    boolean_not_equal = P(8, 'boolean not equal', ApplicationRoleConnectionValueType.boolean)



__all__ = ('EMBED_UPDATE_EMBED_ADD', 'EMBED_UPDATE_EMBED_REMOVE', 'EMBED_UPDATE_NONE', 'EMBED_UPDATE_SIZE_UPDATE',)

CONTENT_LENGTH_MAX = 4000
NONCE_LENGTH_MAX = 25


EMBED_UPDATE_NONE = 0
EMBED_UPDATE_SIZE_UPDATE = 1
EMBED_UPDATE_EMBED_ADD = 2
EMBED_UPDATE_EMBED_REMOVE = 3


MESSAGE_STATE_MASK_TEMPLATE = 1 << 0
MESSAGE_STATE_MASK_SHOULD_UPDATE = 1 << 1
MESSAGE_STATE_MASK_DELETED = 1 << 2

MESSAGE_STATE_MASK_CACHE_MENTIONED_CHANNELS = 1 << 3

MESSAGE_STATE_MASK_PARTIAL_ALL = (
    MESSAGE_STATE_MASK_TEMPLATE | MESSAGE_STATE_MASK_SHOULD_UPDATE | MESSAGE_STATE_MASK_DELETED
)

MESSAGE_STATE_MASK_CACHE_ALL = MESSAGE_STATE_MASK_CACHE_MENTIONED_CHANNELS



__all__ = ('GuildJoinRequestStatus',)

from scarletio import copy_docs

from ...bases import Preinstance as P, PreinstancedBase


class GuildJoinRequestStatus(PreinstancedBase, value_type = str):
    """
    Represents the status of a ``GuildJoinRequest``.
    
    Attributes
    ----------
    name : `str`
        The name of the guild join request status.
    
    value : `str`
        The Discord side identifier value of the guild join request status.
    
    Type Attributes
    ---------------
    Every predefined guild join request status can be accessed as type attribute as well:
    
    +-----------------------+-----------+-----------+
    | Type attribute name   | Name      | Value     |
    +=======================+===========+===========+
    | approved              | approved  | APPROVED  |
    +-----------------------+-----------+-----------+
    | pending               | pending   | PENDING   |
    +-----------------------+-----------+-----------+
    | rejected              | rejected  | REJECTED  |
    +-----------------------+-----------+-----------+
    | started               | started   | STARTED   |
    +-----------------------+-----------+-----------+
    """
    __slots__ = ()
    
    @copy_docs(PreinstancedBase.__new__)
    def __new__(cls, value, name = None):
        if name is None:
            name = value.lower().replace('_', ' ')
        
        return PreinstancedBase.__new__(cls, value, name)
    
    approved = P('APPROVED', 'approved')
    pending = P('PENDING', 'pending')
    rejected = P('REJECTED', 'rejected')
    started = P('STARTED', 'started')



__all__ = ()

AUTO_ARCHIVE_DEFAULT = 3600
AUTO_ARCHIVE_OPTIONS = frozenset((3600, 86400, 259200, 604800))

BITRATE_MIN = 8000
BITRATE_MAX = 384000
BITRATE_DEFAULT = 48000

NAME_LENGTH_MIN = 1
NAME_LENGTH_MAX = 100
NAME_DEFAULT = 'channel'

SLOWMODE_MIN = 0
SLOWMODE_MAX = 21600
SLOWMODE_DEFAULT = 0

STATUS_LENGTH_MIN = 0
STATUS_LENGTH_MAX = 500

TOPIC_LENGTH_MIN = 0
TOPIC_LENGTH_MAX = 1024

USER_LIMIT_MIN = 0
USER_LIMIT_MAX = 99
USER_LIMIT_DEFAULT = 0



__all__ = ()

from ...field_parsers import preinstanced_parser_factory
from ...field_putters import preinstanced_optional_putter_factory
from ...field_validators import preinstanced_validator_factory

from .preinstanced import SessionPlatformType, Status


# desktop

parse_desktop = preinstanced_parser_factory('desktop', Status, Status.offline)
put_desktop = preinstanced_optional_putter_factory('desktop', Status.offline)
validate_desktop = preinstanced_validator_factory('desktop', Status)


# embedded

parse_embedded = preinstanced_parser_factory('embedded', Status, Status.offline)
put_embedded = preinstanced_optional_putter_factory('embedded', Status.offline)
validate_embedded = preinstanced_validator_factory('embedded', Status)


# mobile

parse_mobile = preinstanced_parser_factory('mobile', Status, Status.offline)
put_mobile = preinstanced_optional_putter_factory('mobile', Status.offline)
validate_mobile = preinstanced_validator_factory('mobile', Status)


# platform

validate_platform = preinstanced_validator_factory('platform', SessionPlatformType)


# web

parse_web = preinstanced_parser_factory('web', Status, Status.offline)
put_web = preinstanced_optional_putter_factory('web', Status.offline)
validate_web = preinstanced_validator_factory('web', Status)



__all__ = ()

import sys

from .plugin_finder import PluginFinder


def set_spec_finder():
    """
    Sets``PluginFinder`` into `sys.meta_path`. If it is already set, replaces it.
    """
    meta_path = getattr(sys, 'meta_path', None)
    if (meta_path is None) or (not isinstance(meta_path, list)):
        sys.meta_path = meta_path = []
    
    for index in range(len(meta_path)):
        path_finder = meta_path[index]
        if type(path_finder).__name__ == PluginFinder.__name__:
            break
    
    else:
        index = -1
    
    plugin_finder = PluginFinder()
    
    if index == -1:
        meta_path.insert(0, plugin_finder)
    else:
        meta_path[index] = plugin_finder



__all__ = ('GuildActivityOverviewActivity',)

from scarletio import RichAttributeErrorBaseType

from .fields import parse_level, parse_score, put_level, put_score, validate_level, validate_score
from .preinstanced import GuildActivityOverviewActivityLevel


class GuildActivityOverviewActivity(RichAttributeErrorBaseType):
    """
    Activity information inside of a guild activity overview.
    
    Attributes
    ----------
    level : ``GuildActivityOverviewActivityLevel``
        The activity's level.
    
    score : `int`
        Total accumulated score.
    """
    __slots__ = ('level', 'score')
    
    def __new__(cls, *, level = ..., score = ...):
        """
        Creates a new guild activity overview activity.
        
        Parameters
        ----------
        level : `None | GuildActivityOverviewActivityLevel | int`, Optional (Keyword only)
            The activity's level.
        
        score : `int`, Optional (Keyword only)
            Total accumulated score.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value in incorrect.
        """
        # level
        if level is ...:
            level = GuildActivityOverviewActivityLevel.none
        else:
            level = validate_level(level)
        
        # score
        if score is ...:
            score = 0
        else:
            score = validate_score(score)
        
        # Construct
        self = object.__new__(cls)
        self.level = level
        self.score = score
        return self
    
    
    @classmethod
    def _create_empty(cls):
        """
        Creates an empty activity overview activity.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.level = GuildActivityOverviewActivityLevel.none
        self.score = 0
        return self
    
    
    def __repr__(self):
        """Returns repr(self)."""
        repr_parts = ['<', type(self).__name__]
        
        # level
        level = self.level
        repr_parts.append(' level = ')
        repr_parts.append(level.name)
        repr_parts.append(' ~ ')
        repr_parts.append(repr(level.value))
        
        # score
        repr_parts.append(', score = ')
        repr_parts.append(repr(self.score))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns hash(self)."""
        hash_value = 0
        
        # level
        hash_value ^= hash(self.level)
        
        # score
        hash_value ^= self.score
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns self == other."""
        if type(self) is not type(other):
            return NotImplemented
        
        # level
        if self.level is not other.level:
            return False
        
        # score
        if self.score != other.score:
            return False
        
        return True
        
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new activity information.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Channel data receive from Discord.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.level = parse_level(data)
        self.score = parse_score(data)
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Serializes the guild activity overview activity.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default field values should be included.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        put_level(self.level, data, defaults)
        put_score(self.score, data, defaults)
        return data
    
    
    def copy(self):
        """
        Copies the guild activity overview activity.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.level = self.level
        new.score = self.score
        return new
    
    
    def copy_with(self, *, level = ..., score = ...):
        """
        Copies the guild activity overview activity with the given fields.
        
        Parameters
        ----------
        level : `None | GuildActivityOverviewActivityLevel | int`, Optional (Keyword only)
            The activity's level.
        
        score : `int`, Optional (Keyword only)
            Total accumulated score.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value in incorrect.
        """
        # level
        if level is ...:
            level = self.level
        else:
            level = validate_level(level)
        
        # score
        if score is ...:
            score = self.score
        else:
            score = validate_score(score)
        
        # Construct
        new = object.__new__(type(self))
        new.level = level
        new.score = score
        return new



__all__ = ('ApplicationInstallParameters',)

from scarletio import RichAttributeErrorBaseType

from ...permission import Permission

from .fields import (
    parse_permissions, parse_scopes, put_permissions, put_scopes, validate_permissions, validate_scopes
)


class ApplicationInstallParameters(RichAttributeErrorBaseType):
    """
    Parameters for inviting a bot.
    
    Attributes
    ----------
    permissions : ``Permission``
        The permissions to invite the bot with.
    scopes : `None | tuple<str>`
        Oauth2 scopes to invite the bot with.
    """
    __slots__ = ('permissions', 'scopes')
    
    def __new__(cls, *, permissions = ..., scopes = ...):
        """
        Creates a new application install parameters.
        
        Parameters
        ----------
        permissions : `int`, ``Permission``, Optional (Keyword only)
            The permissions to invite the bot with.
        
        scopes : `None`, `iterable` of (``Oauth2Scope``, `str`), Optional (Keyword only)
            Oauth2 scopes to invite the bot with.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        if permissions is ...:
            permissions = Permission()
        else:
            permissions = validate_permissions(permissions)
        
        if scopes is ...:
            scopes = None
        else:
            scopes = validate_scopes(scopes)
        
        self = object.__new__(cls)
        self.permissions = permissions
        self.scopes = scopes
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a application installation parameters instance from the given data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Application installation parameters data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.scopes = parse_scopes(data)
        self.permissions = parse_permissions(data)
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the application install parameters to json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether fields with their default value should be included as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        put_permissions(self.permissions, data, defaults)
        put_scopes(self.scopes, data, defaults)
        return data
    
    
    def __repr__(self):
        """Returns the application install parameters' representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' scopes = [')
        
        scopes = self.scopes
        if (scopes is not None):
            length = len(scopes)
            index = 0
            
            while True:
                scope = scopes[index]
                repr_parts.append(repr(scope))
                
                index += 1
                if index == length:
                    break
                
                repr_parts.append(', ')
                continue
        
        repr_parts.append(']')
        
        repr_parts.append(', permissions = ')
        repr_parts.append(format(self.permissions, 'd'))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns the application install parameters' hash value."""
        hash_value = 0
        
        scopes = self.scopes
        if (scopes is not None):
            hash_value ^= len(scopes) << 16
            
            for scope in scopes:
                hash_value ^= hash(scope)
        
        hash_value ^= self.permissions
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two application install parameters are equal."""
        if type(other) is not ApplicationInstallParameters:
            return NotImplemented
        
        if self.permissions != other.permissions:
            return False
        
        if self.scopes != other.scopes:
            return False
        
        return True
    
    
    def copy(self):
        """
        Copies the application install parameters.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.permissions = self.permissions
        scopes = self.scopes
        if (scopes is not None):
            scopes = (*scopes, )
        new.scopes = scopes
        return new

    
    def copy_with(self, permissions = ..., scopes = ...):
        """
        Copies the application install parameters with the given fields.
        
        Parameters
        ----------
        permissions : `int`, ``Permission``, Optional (Keyword only)
            The permissions to invite the bot with.
        
        scopes : `None`, `iterable` of (``Oauth2Scope``, `str`), Optional (Keyword only)
            Oauth2 scopes to invite the bot with.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        if permissions is ...:
            permissions = self.permissions
        else:
            permissions = validate_permissions(permissions)
        
        if scopes is ...:
            scopes = self.scopes
            if (scopes is not None):
                scopes = (*scopes, )
        else:
            scopes = validate_scopes(scopes)
        
        new = object.__new__(type(self))
        new.permissions = permissions
        new.scopes = scopes
        return new



__all__ = ()

from scarletio import RichAttributeErrorBaseType

from ..constants import PLUGIN_ACTION_FLAG_LOAD


class PluginTreeIterator(RichAttributeErrorBaseType):
    """
    Plugin tree iterator iterating over their plugins.
    
    Attributes
    ----------        
    action : `int`
        Plugin loader action flag that will be executed on the plugins.
    current_plugin_tree : `None`, ``PluginTree``
        The plugin tree that's plugins are currently iterated over.
    current_plugins : `None`, `list` of ``Plugin``
        The plugins we are currently iterating over. They belong to ``.current_plugin_tree``.
    done_cancelled : `None`, `set` of ``PluginTree``
        Cancelled plugin trees.
    done_fail : `None`, `set` of ``PluginTree``
        Failed plugin trees.
    done_success : `None`, `set` of ``PluginTree``
        Plugin trees passed successfully.
    iterated : `None`, `set` of ``Plugin``
        Already iterated plugins.
    to_do : `None`, `list` of ``PluginTree``
       Plugin trees sorted in reversed order to the action executed. (Always the last one is popped.)
    """
    __slots__ = (
        'action', 'current_plugin_tree', 'current_plugins', 'done_cancelled', 'done_fail', 'done_success',
        'iterated', 'to_do'
    )
    
    def __new__(cls, plugin_trees, action):
        """
        Creates a new plugin tree iterator.
        
        Parameters
        ----------
        plugin_trees : `list` of ``PluginTree``
            Already sorted plugin trees.
        
        action : `int`
            Plugin loader action flag that will be executed on the plugins.
        """
        to_do = plugin_trees.copy()
        
        if action == PLUGIN_ACTION_FLAG_LOAD:
            to_do.reverse()
        
        if not to_do:
            to_do = None
        
        self = object.__new__(cls)
        self.action = action
        self.current_plugin_tree = None
        self.current_plugins = None
        self.done_success = None
        self.done_fail = None
        self.done_cancelled = None
        self.iterated = None
        self.to_do = to_do
        return self
    
    
    def __iter__(self):
        """
        Returns an iterator over self.
        
        Returns
        -------
        self : `instance<type<self>>`
        """
        return self
    
    
    def __next__(self):
        """
        Gets the next element of self.
        
        Returns
        -------
        plugin : ``Plugin``
        
        Raises
        ------
        StopIteration
            - If there are no more plugins to iterate trough.
        """
        while True:
            next_plugin = self._try_set_next_plugin_tree_and_pull()
            if (next_plugin is None):
                break
            
            if self._register_as_iterated(next_plugin):
                return next_plugin
        
        raise StopIteration()
    
    
    def _try_set_next_plugin_tree_and_pull(self):
        """
        Tries to set the next plugin tree if required and pulls a new plugin.
        
        Returns
        -------
        plugin : `None`, ``Plugin``
        """
        current_plugins = self.current_plugins
        current_plugin_tree = self.current_plugin_tree
        # Iteration over perhaps?
        if (current_plugins is None):
            if (current_plugin_tree is not None):
                self.add_done_success(current_plugin_tree)
            
            return self._set_next_plugin_tree_and_pull()
        
        if current_plugin_tree is not None:
            return self._pull_next_plugin()
        
        return self._set_next_plugin_tree_and_pull()
    
    
    def _set_next_plugin_tree_and_pull(self):
        """
        Sets the next plugin plugin tree and pulls a new plugin.
        
        Returns
        -------
        plugin : `None`, ``Plugin``
        """
        while True:
            to_do = self.to_do
            if (to_do is None):
                return None
            
            current_plugin_tree = to_do.pop()
            if not to_do:
                self.to_do = None
            
            current_plugins = current_plugin_tree.get_plugins_sorted()
            if not current_plugins:
                self.add_done_success(current_plugin_tree)
                continue
            
            current_plugins = current_plugins.copy()
            
            if self.action == PLUGIN_ACTION_FLAG_LOAD:
                current_plugins.reverse()
            
            self.current_plugin_tree = current_plugin_tree
            self.current_plugins = current_plugins
            return self._pull_next_plugin()
    
    
    def _pull_next_plugin(self):
        """
        Pulls the next plugin from the currently set plugin.
        
        Returns
        -------
        plugin : `None`, ``Plugin``
        """
        current_plugins = self.current_plugins
        if (current_plugins is None):
            return None
        
        next_plugin = current_plugins.pop()
        if not current_plugins:
            self.current_plugins = None
        
        return next_plugin
    
    
    def add_done_success(self, plugin_tree):
        """
        Adds a done plugin tree as successfully iterated. The plugin must be not
        
        Parameters
        ----------
        plugin_tree : ``PluginTree``
        """
        if self.has_done(plugin_tree):
            return
        
        done_success = self.done_success
        if (done_success is None):
            done_success = set()
            self.done_success = done_success
        
        done_success.add(plugin_tree)
    
    
    def add_done_fail(self, plugin_tree):
        """
        Adds a done plugin tree as failed.
        
        Parameters
        ----------
        plugin_tree : ``PluginTree``
        """
        done_fail = self.done_fail
        if (done_fail is None):
            done_fail = set()
            self.done_fail = done_fail
        
        done_fail.add(plugin_tree)
    
    
    def add_done_cancelled(self, plugin_tree):
        """
        Adds a done plugin tree as cancelled.
        
        Parameters
        ----------
        plugin_tree : ``PluginTree``
        """
        done_cancelled = self.done_cancelled
        if (done_cancelled is None):
            done_cancelled = set()
            self.done_cancelled = done_cancelled
        
        done_cancelled.add(plugin_tree)
    
    
    def has_done(self, plugin_tree):
        """
        Returns whether the given plugin tree is already in any done groups of the iterator.
        
        Parameters
        ----------
        plugin_tree : ``PluginTree``
            The plugin tree to check for.
        
        Returns
        -------
        has_done : `bool`
        """
        done_success = self.done_success
        if (done_success is not None) and (plugin_tree in done_success):
            return True
        
        done_fail = self.done_fail
        if (done_fail is not None) and (plugin_tree in done_fail):
            return True
    
        done_cancelled = self.done_cancelled
        if (done_cancelled is not None) and (plugin_tree in done_cancelled):
            return True
        
        return False
    
    
    def _register_as_iterated(self, plugin):
        """
        Registers the given plugin as iterated.
        
        Parameters
        ----------
        plugin : ``Plugin``
            the plugin to register.
        
        Returns
        -------
        freshly_registered : `bool`
            Whether the plugin was not registered before.
        """
        iterated = self.iterated
        if (iterated is None):
            iterated = set()
            self.iterated = iterated
        
        length_old = len(iterated)
        iterated.add(plugin)
        length_new = len(iterated)
        
        return length_old != length_new
    
    
    def fail_current_hard(self):
        """
        Fails the currently iterated plugin tree and removes its plugins from the iteration.
        """
        current_plugin_tree = self.current_plugin_tree
        if (current_plugin_tree is not None):
            self.add_done_fail(current_plugin_tree)
            self.current_plugin_tree = None
            self.current_plugins = None
    
    
    def fail_current_soft(self):
        """
        Fails the currently iterated plugin tree.
        """
        current_plugin_tree = self.current_plugin_tree
        if (current_plugin_tree is not None):
            self.add_done_fail(current_plugin_tree)
        
    
    def cancel_dependents_in_to_do(self, plugins):
        """
        Cancels all plugin trees that contain any of the given plugins.
        
        Parameters
        ----------
        plugins : `set` of ``Plugins``
            The plugins to look for the dependent trees.
        """
        new_to_do = None
        old_to_do = self.to_do
        
        while (old_to_do is not None):
            plugin_tree = old_to_do.pop()
            if not old_to_do:
                old_to_do = None
            
            if plugin_tree.has_intersection_with_plugins(plugins):
                if new_to_do is None:
                    new_to_do = []
                
                new_to_do.append(plugin_tree)
            
            else:
                self.add_done_cancelled(plugin_tree)
        
        if (new_to_do is not None):
            new_to_do.reverse()
        self.to_do = new_to_do
    
    
    def iter_done_success(self):
        """
        Iterates over the done (successfully) plugin trees of the plugin tree iterator.
        
        This method is an iterable generator.
        
        Yields
        ------
        plugin_tree : ``PluginTree``
        """
        done_success = self.done_success
        if (done_success is not None):
            yield from done_success


    def iter_done_cancelled(self):
        """
        Iterates over the done (cancelled) plugin trees of the plugin tree iterator.
        
        This method is an iterable generator.
        
        Yields
        ------
        plugin_tree : ``PluginTree``
        """
        done_cancelled = self.done_cancelled
        if (done_cancelled is not None):
            yield from done_cancelled
    
    
    def iter_done_fail(self):
        """
        Iterates over the done (failed) plugin trees of the plugin tree iterator.
        
        This method is an iterable generator.
        
        Yields
        ------
        plugin_tree : ``PluginTree``
        """
        done_fail = self.done_fail
        if (done_fail is not None):
            yield from done_fail
